		
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <malloc.h>
#define TRUE 1
#define FALSE 0
#define MAX 100
#define MAX_DIMENSIONS 15
#define PACKET_MAX 500
#define END 0
#define NAME 1
#define QUALIFIER 2
#define TYPE 3
#define NUMBER 4
#define IN 5
#define OUT 6
#define INOUT 7
#define UNKNOWN 99
typedef int bool ;

struct ArgNode
{	char ArgName[MAX] ;
	char ArgType[MAX] ;
	char Qualifier[MAX] ;
	int ArgDim[MAX_DIMENSIONS] ;
	int ArgNumDim ;
	int ArgMode ;
	struct ArgNode *Next ;
} ;

struct ArgList
{	struct ArgNode *Head ;
	struct ArgNode *Current ;
} ;

struct FnStruct
{       char FnName[MAX] ;
	struct ArgList Args ;
	char RetType[MAX] ;
} ;

void InitFnStruct(struct FnStruct *F)
{	strcpy(F->FnName,"") ;
	(F->Args).Head = (F->Args).Current = NULL ;
	strcpy(F->RetType,"") ;
}

struct FnList
{	struct FnStruct *Fn ;
	struct FnList *Next ;
} ;

void MoveFirst(struct ArgList *L)
{	L->Current = L->Head ;}

struct ArgNode *Search(struct ArgList *L,char *ArgNm)
{	struct ArgNode *P = L->Head ;
	while(P != NULL)
	{	if(strcmp(P->ArgName,ArgNm) == 0)
			break ;
		P = P->Next ;
	}
	return P ;
}

struct FnList *FnSearch(struct FnList *L,char *FnName)
{	struct FnList *P = L ;
	while(P != NULL)
	{	if(strcmp((P->Fn)->FnName,FnName) == 0)
			break ;
		P = P->Next ;
	}
	return P ;
}

void MoveLast(struct ArgList *L)
{	L->Current = L->Head ;
	while((L->Current)->Next)
		L->Current = (L->Current)->Next ;
}

struct ISLScanner
{       char *Str ;
	int Pos ;
	int Len ;
	char TokenStr[MAX] ;
} ;

void PrintError(struct ISLScanner *S,const char *ErrMsg)
{	int i ;
	printf("\n%s\n",S->Str) ;
	for(i = 0 ; i < S->Pos ; i++)
		printf(" ") ;
	printf("*") ;
	printf("\n%s\n",ErrMsg) ;
}

void InitISLScanner(struct ISLScanner *S,char *ParseStr)
{	S->Str = (char *)malloc(MAX) ;
	strcpy(S->Str,ParseStr) ;
	S->Pos = S->Len = 0 ;
	strcpy(S->TokenStr,"") ;
}

int DeTokenize(struct ISLScanner *S)
{	S->Pos -= S->Len ;
	return S->Len ;
}

int Tokenize(struct ISLScanner *S)
{	int State = 0 ;
	int i = 0 ;
	while(S->Str[S->Pos] == ' ' || S->Str[S->Pos] == '\t')
		S->Pos++ ;
	while(1)
	{       switch(State)
		{	case 0 :
				if(S->Str[S->Pos] == '\0')
				{	S->Len = 0 ;
					S->TokenStr[0] = '\0' ;
					return END ;
				}
				if((S->Str[S->Pos] >= 'A' && S->Str[S->Pos] <= 'Z') ||
				   (S->Str[S->Pos] >= 'a' && S->Str[S->Pos] <= 'z'))
				{	S->TokenStr[i++] = S->Str[S->Pos] ;
					State = 1 ;
				}
				else if(S->Str[S->Pos] >= '0' && S->Str[S->Pos] <= '9')
				{	S->TokenStr[i++] = S->Str[S->Pos] ;
					State = 2 ;
				}
				else if(S->Str[S->Pos] == '<' && S->Str[S->Pos+1] == '-')
				{       S->TokenStr[0] = S->Str[S->Pos] ;
					S->TokenStr[1] = S->Str[S->Pos+1] ;
					S->Pos += 2 ;
					S->Len = 2 ;
					if(S->Str[S->Pos] == '>')
					{       S->TokenStr[2] = S->Str[S->Pos] ;
						S->TokenStr[3] = '\0' ;
						S->Pos++ ;
						S->Len++ ;
						return INOUT ;
					}
					else
					{	S->TokenStr[2] = '\0' ;
						return IN ;
					}
				}
				else if(S->Str[S->Pos] == '-' && S->Str[S->Pos+1] == '>')
				{	S->TokenStr[0] = S->Str[S->Pos] ;
					S->TokenStr[1] = S->Str[S->Pos+1] ;
					S->TokenStr[2] = '\0' ;
					S->Pos += 2 ;
					S->Len = 2 ;
					return OUT ;
				}
				else
				{	switch(S->Str[S->Pos])
					{	case ',' :
						case '(' :
						case ')' :
						case '*' :
						case '[' :
						case ']' :
							S->Pos++ ;
							S->TokenStr[0] = S->Str[S->Pos] ;
							S->Len = 1 ;
							S->TokenStr[S->Len] = '\0' ;
							return S->Str[S->Pos-1] ;
					}
					S->Pos++ ;
					S->Len = 1 ;
					return UNKNOWN ;
				}
				break ;
			case 1 :
				if((S->Str[S->Pos] >= 'A' && S->Str[S->Pos] <= 'Z') ||
				   (S->Str[S->Pos] >= 'a' && S->Str[S->Pos] <= 'z') ||
				   (S->Str[S->Pos] >= '0' && S->Str[S->Pos] <= '9') ||
				   S->Str[S->Pos] == '_')
				{	S->TokenStr[i++] = S->Str[S->Pos] ;
					State = 1 ;
				}
				else
				{	S->TokenStr[i] = '\0' ;
					S->Len = strlen(S->TokenStr) ;
					if(strcmp(S->TokenStr,"int") == 0 ||
					   strcmp(S->TokenStr,"char") == 0 ||
					   strcmp(S->TokenStr,"float") == 0 ||
					   strcmp(S->TokenStr,"double") == 0)
						return TYPE ;
					else if(strcmp(S->TokenStr,"long") == 0 ||
						strcmp(S->TokenStr,"short") == 0 ||
						strcmp(S->TokenStr,"unsigned") == 0 ||
						strcmp(S->TokenStr,"signed") == 0)
						return QUALIFIER ;
					else
						return NAME ;
				}
				break ;
			case 2 :
				if(S->Str[S->Pos] >= '0' && S->Str[S->Pos] <= '9')
				{	S->TokenStr[i++] = S->Str[S->Pos] ;
					State = 2 ;
				}
				else
				{	S->TokenStr[i] = '\0' ;
					S->Len = strlen(S->TokenStr) ;
					return NUMBER ;
				}
				break ;

		}
		S->Pos++ ;
	}
}

bool Dimensions(struct ISLScanner *S,struct ArgNode *A1)
{       if(Tokenize(S) == '[' && Tokenize(S) == NUMBER)
	{	A1->ArgDim[A1->ArgNumDim++] = atoi(S->TokenStr) ;
		if(Tokenize(S) == ']')
		{	if(Tokenize(S) == '[')
			{	DeTokenize(S) ;
				if(Dimensions(S,A1))
					return TRUE ;
				else
					return FALSE ;
			}
			else
			{	DeTokenize(S) ;
				return TRUE ;
			}
		}
		else
		{	DeTokenize(S) ;
			PrintError(S,"Expected ]") ;
			return FALSE ;
		}
	}
	else
	{	DeTokenize(S) ;
		PrintError(S,"Expected number") ;
		return FALSE ;
	}
}

bool Arg(struct ISLScanner *S,struct ArgList *A1)
{       int Token,Mode ;
	(A1->Current)->Next = NULL ;
	if(Tokenize(S) == NAME)
	{       strcpy((A1->Current)->ArgName,S->TokenStr) ;
		if(Search(A1,(A1->Current)->ArgName) != A1->Current)
		{       DeTokenize(S) ;
			PrintError(S,"Argument name repeated") ;
			return FALSE ;
		}
		Token = Tokenize(S) ;
		if(Token == TYPE)
		{       strcpy((A1->Current)->ArgType,S->TokenStr) ;
			strcpy((A1->Current)->Qualifier,"") ;
			if(Tokenize(S) == '[')
			{	DeTokenize(S) ;
				if(Dimensions(S,(A1->Current)))
				{	Mode = Tokenize(S) ;
					if(Mode == OUT || Mode == IN || Mode == INOUT)
					{	(A1->Current)->ArgMode = Mode ;
						return TRUE ;
					}
					else
						return FALSE ;
				}
				else
					return FALSE ;
			}
			else
			{	DeTokenize(S) ;
				Mode = Tokenize(S) ;
				if(Mode == OUT || Mode == IN || Mode == INOUT)
				{	(A1->Current)->ArgMode = Mode ;
					return TRUE ;
				}
				else
				{       DeTokenize(S) ;
					PrintError(S,"Expected argument mode <- IN / -> OUT / <-> INOUT") ;
					return FALSE ;
				}
			}
		}
		else if(Token == QUALIFIER)
		{	strcpy((A1->Current)->Qualifier,S->TokenStr) ;
			if(Tokenize(S) == TYPE)
			{       strcpy((A1->Current)->ArgType,S->TokenStr) ;
				if(Tokenize(S) == '[')
				{	DeTokenize(S) ;
					if(Dimensions(S,(A1->Current)))
					{	Mode = Tokenize(S) ;
						if(Mode == OUT || Mode == IN || Mode == INOUT)
						{	(A1->Current)->ArgMode = Mode ;
							return TRUE ;
						}
						else
						{       DeTokenize(S) ;
PrintError(S,"Expected argument mode <- IN / -> OUT / <-> INOUT") ;
							return FALSE ;
						}
					}
					else
						return FALSE ;
				}
				else
				{	DeTokenize(S) ;
					Mode = Tokenize(S) ;
					if(Mode == OUT || Mode == IN || Mode == INOUT)
					{	(A1->Current)->ArgMode = Mode ;
						return TRUE ;
					}
					else
					{       DeTokenize(S) ;
						PrintError(S,"Expected argument mode <- IN / -> OUT / <-> INOUT") ;
						return FALSE ;
					}
				}
			}
			else
			{	DeTokenize(S) ;
				strcpy((A1->Current)->ArgType,"") ;
				if(Tokenize(S) == '[')
				{	DeTokenize(S) ;
					if(Dimensions(S,(A1->Current)))
					{	Mode = Tokenize(S) ;
						if(Mode == OUT || Mode == IN || Mode == INOUT)
						{	(A1->Current)->ArgMode = Mode ;
							return TRUE ;
						}
						else
						{       DeTokenize(S) ;
							PrintError(S,"Expected argument mode <- IN / -> OUT / <-> INOUT") ;
							return FALSE ;
						}
					}
					else
						return FALSE ;
				}
				else
				{	DeTokenize(S) ;
					Mode = Tokenize(S) ;
					if(Mode == OUT || Mode == IN || Mode == INOUT)
					{	(A1->Current)->ArgMode = Mode ;
						return TRUE ;
					}
					else
					{       DeTokenize(S) ;
						PrintError(S,"Expected argument mode <- IN / -> OUT / <-> INOUT") ;
						return FALSE ;
					}
				}
			}
		}
		else
		{       DeTokenize(S) ;
			PrintError(S,"Expected argument type") ;
			return FALSE ;
		}
	}
	else
	{       DeTokenize(S) ;
		PrintError(S,"Expected argument name") ;
		return FALSE ;
	}
}

bool ArgListing(struct ISLScanner *S,struct ArgList *L1)
{       if(L1->Head == NULL)
		L1->Head = L1->Current = (struct ArgNode *)malloc(sizeof(struct ArgNode)) ;
	else
	{	MoveLast(L1) ;
		(L1->Current)->Next = (struct ArgNode *)malloc(sizeof(struct ArgNode)) ;
		L1->Current = (L1->Current)->Next ;
	}
	if(Arg(S,L1))
	{       (L1->Current)->Next = NULL ;
		if(Tokenize(S) == ',')
			if(ArgListing(S,L1))
				return TRUE ;
			else
				return FALSE ;
		else
		{	DeTokenize(S) ;
			return TRUE ;
		}
	}
	else
	{	L1->Current = NULL ;
		return FALSE ;
	}
}

bool FindReturnVal(struct ISLScanner *S,struct FnStruct *F1)
{	int Token = Tokenize(S) ;
	if(Token != END && Token != UNKNOWN)
		strcpy(F1->RetType,S->TokenStr) ;
	else if(Token == UNKNOWN)
	{	DeTokenize(S) ;	
		PrintError(S,"Expected function return type") ;
		return FALSE ;
	}
	else
		strcpy(F1->RetType,"int") ;
	return TRUE ;
}

bool Function(struct ISLScanner *S,struct FnStruct *Fn)
{	if(Tokenize(S) == NAME)
	{	strcpy(Fn->FnName,S->TokenStr) ;
		if(Tokenize(S) == '(')
			if(Tokenize(S) == ')')
			{	if(FindReturnVal(S,Fn))
					return TRUE ;
				else
					return FALSE ;
			}
			else
			{	DeTokenize(S) ;
				if(ArgListing(S,&Fn->Args))
					if(Tokenize(S) == ')')
					{	if(FindReturnVal(S,Fn))
							return TRUE ;
						else
							return FALSE ;
					}
					else
					{       DeTokenize(S) ;
						PrintError(S,"Expected end of function") ;
						return FALSE ;
					}
				else
					return FALSE ;
			}
		else
		{	DeTokenize(S) ;
			PrintError(S,"Expected (") ;
			return FALSE ;
		}
	}
	else
	{       DeTokenize(S) ;
		PrintError(S,"Expected function name") ;
		return FALSE ;
	}
}

int WriteArgname(FILE *Fp,struct FnStruct *Fn)
{	struct ArgNode *Temp;
	int i ;
	Temp=(Fn->Args).Head;
	fprintf(Fp,"char *sendaddr");
	while(Temp!=NULL)
	{	fprintf(Fp,",%s %s",Temp->Qualifier,Temp->ArgType);
		fprintf(Fp," ");
		if((Temp->ArgMode == OUT || Temp->ArgMode == INOUT) && Temp->ArgNumDim == 0)
			fprintf(Fp,"*%s",Temp->ArgName);
		else			
			fprintf(Fp,"%s",Temp->ArgName);
		if(Temp->ArgNumDim != 0)
		{	for(i = 0 ; i < Temp->ArgNumDim ; i++)
				fprintf(Fp,"[%d]",Temp->ArgDim[i]);
		}
		Temp=Temp->Next;
	}
	return 0;
}

int WriteFunctionname(FILE *Fp,struct FnStruct *Fn)
{	char *Temp;
	fprintf(Fp,"%s ",Fn->RetType);
	Temp=(char *)malloc((strlen(Fn->FnName)+6)*sizeof(char));
	strcpy(Temp,Fn->FnName);
	strcat(Temp,"_rpc");
	fprintf(Fp,Temp);
	free(Temp);
	return 1;
}

int WriteArgcp(FILE *Fp,struct FnStruct *Fn)
{	struct ArgNode *Temp= (Fn->Args).Head ;
	while(Temp!=NULL)
	{	if(Temp->ArgMode == IN || Temp->ArgMode == INOUT)
			if(Temp->ArgNumDim != 0)
			{	int Size = 1,i ;	
				for(i = 0 ; i < Temp->ArgNumDim ; i++)
					Size *= Temp->ArgDim[i] ;
				fprintf(Fp,"\n\tcopy_to_buf(BUF,(char *)%s,sizeof(%s %s)*%d,&index);",Temp->ArgName,Temp->Qualifier,Temp->ArgType,Size);
			}
			else
				if(Temp->ArgMode == IN)
					fprintf(Fp,"\n\tcopy_to_buf(BUF,(char *)&%s,sizeof(%s %s),&index);",Temp->ArgName,Temp->Qualifier,Temp->ArgType);
				else
					fprintf(Fp,"\n\tcopy_to_buf(BUF,(char *)%s,sizeof(%s %s),&index);",Temp->ArgName,Temp->Qualifier,Temp->ArgType);		
		Temp=Temp->Next;
	}
	return 1 ;
}

int WriteArgcp2(FILE *Fp,struct FnStruct *Fn)
{	struct ArgNode *Temp=(Fn->Args).Head ;
	while(Temp!=NULL)
	{	if(Temp->ArgMode == IN || Temp->ArgMode == INOUT)
			if(Temp->ArgNumDim != 0)
			{	int Size = 1,i ;	
				for(i = 0 ; i < Temp->ArgNumDim ; i++)
					Size *= Temp->ArgDim[i] ;
				fprintf(Fp,"\n\tcopy_from_buf(BUF,(char *)%s,sizeof(%s %s)*%d,index);",Temp->ArgName,Temp->Qualifier,Temp->ArgType,Size);
			}
			else
				fprintf(Fp,"\n\tcopy_from_buf(BUF,(char *)&%s,sizeof(%s %s),index);",Temp->ArgName,Temp->Qualifier,Temp->ArgType);
		Temp=Temp->Next;
	}
	return 1 ;
}

int WriteCopies(FILE *Fp){
	fprintf(Fp,"int copy_from_buf(char *buf,char *aname,int size,int *index){");
	fprintf(Fp,"\n\tint i;");
	fprintf(Fp,"\n\tfor(i=0;i<(size);i++){");
	fprintf(Fp,"\n\t\taname[i]=buf[*index];");
	fprintf(Fp,"\n\t\t*index=*index+1;\n\t}");
	fprintf(Fp,"\n}");
	fprintf(Fp,"\nint copy_to_buf(char *buf,char *aname,int size,int *index){");
	fprintf(Fp,"\n\tint i;");
	fprintf(Fp,"\n\tfor(i=0;i<(size);i++){");
	fprintf(Fp,"\n\t\tbuf[*index]=aname[i];");
	fprintf(Fp,"\n\t\t*index=*index+1;\n\t}");
	fprintf(Fp,"\n\tbuf[*index]=\'\\0\';");
	fprintf(Fp,"\n}\n");
	return 1;

}

int CreateClient(char *Fn,struct FnList *FL)
{	char *Fname;
	char *Tmp;
	FILE *Fp;
	int Pid=1;
	struct FnList *Top = FL ;
	struct ArgNode *Temp ;
	Fname=(char *)malloc(sizeof(char)*strlen(Fn)+7);
	strcpy(Fname,Fn);
	if((Tmp=strstr(Fname,"."))!=NULL)
		Tmp[0]='\0';
	strcat(Fname,"_cli.c");
	if((Fp=fopen(Fname,"w+"))==NULL)
	{	perror("open");
		free(Fname);
		return(1);
	}
	free(Fname);
	fprintf(Fp,"/*******************************************************") ;
	fprintf(Fp,"\n*RPC client stub file created by RPCGEN");
	fprintf(Fp,"\n*\n*SOURCE IDL FILE : %s",Fn) ;
	fprintf(Fp,"\n*\n*Donot try to edit");
	fprintf(Fp,"\n*******************************************************/\n") ;
	fprintf(Fp,"#include<sys/socket.h>\n");
	fprintf(Fp,"#include<sys/types.h>\n");
	fprintf(Fp,"#include<errno.h>\n");
	fprintf(Fp,"#include<stdlib.h>\n");
	fprintf(Fp,"#include<string.h>\n");
	fprintf(Fp,"#include<netinet/in.h>\n");
	fprintf(Fp,"#include<stdio.h>\n");
	fprintf(Fp,"#include<signal.h>\n");
	fprintf(Fp,"#define TIMEOUT 2\n") ;
	fprintf(Fp,"\n\nvoid TimeoutHandle(int Sig){") ;
	fprintf(Fp,"\n\tprintf(\"\\nTime out expired!\\n\");") ;
	fprintf(Fp,"\n\texit(1);");
	fprintf(Fp,"\n}\n");
	while(Top!=NULL)
	{	WriteFunctionname(Fp,Top->Fn);
		fprintf(Fp,"(");
		WriteArgname(Fp,Top->Fn);
		fprintf(Fp,"){\n");
		fprintf(Fp,"\tchar BUF[1024];");
		fprintf(Fp,"\n\tint index=0;");
		fprintf(Fp,"\n\t%s",(Top->Fn)->RetType);
		fprintf(Fp," retval;\n");
		fprintf(Fp,"\tint sockfd;\n");
		fprintf(Fp,"\tstatic int pr=%d;\n",Pid);
		fprintf(Fp,"\tstruct sockaddr_in their_addr;\n\n");
		fprintf(Fp,"\tsignal(SIGALRM,TimeoutHandle);\n\n") ;
		fprintf(Fp,"/* create a socket */\n");
		fprintf(Fp,"\n\tif((sockfd=socket(AF_INET,SOCK_STREAM,0))==-1) {\n\t\tperror(\"socket\");\n\t\texit(1);\n\t}");
		fprintf(Fp,"\n\ttheir_addr.sin_family=AF_INET;");
		fprintf(Fp,"\n\ttheir_addr.sin_port=htons(3119);");
		fprintf(Fp,"\n\tif(!inet_aton(sendaddr,(struct in_addr *)(&their_addr.sin_addr))){ ");
		fprintf(Fp,"\n\t\tfprintf(stderr,\"Wrong address\");");
		fprintf(Fp,"\n\t\tkill(getpid(),SIGUSR1);");
		fprintf(Fp,"\n\t\treturn(1);\n\t}");
		fprintf(Fp,"\n\tbzero(&(their_addr.sin_zero),8);");
		fprintf(Fp,"\n\tif(connect(sockfd,(struct sockaddr *)&their_addr,sizeof(struct sockaddr))==-1){\n\t\tperror(\"connect\");\n\t\texit(1);\n\t}");
		fprintf(Fp,"\n\tcopy_to_buf(BUF,(char *)&pr,sizeof(int),&index);");
		WriteArgcp(Fp,Top->Fn);
		fprintf(Fp,"\n\tif(send(sockfd,BUF,index,0)==-1){\n\t\tperror(\"send\");\n\t\texit(1);\n\t}");
		fprintf(Fp,"\n\talarm(TIMEOUT);") ;
		fprintf(Fp,"\n\tif(recv(sockfd,BUF,1024,0)==-1){\n\t\tperror(\"recv\");\n\t\texit(1);\n\t}");
		fprintf(Fp,"\n\tindex=0;");
		fprintf(Fp,"\n\tcopy_from_buf(BUF,(char *)&retval,sizeof(retval),&index);\n");
		Temp = ((Top->Fn)->Args).Head ;
		while(Temp != NULL)
		{	if(Temp->ArgMode == OUT || Temp->ArgMode == INOUT)
				if(Temp->ArgNumDim != 0)
				{	int Size = 1,i ;	
					for(i = 0 ; i < Temp->ArgNumDim ; i++)
						Size *= Temp->ArgDim[i] ;
					fprintf(Fp,"\n\tcopy_from_buf(BUF,(char *)%s,sizeof(%s %s)*%d,&index);",Temp->ArgName,Temp->Qualifier,Temp->ArgType,Size);
				}
				else
					fprintf(Fp,"\n\tcopy_from_buf(BUF,(char *)%s,sizeof(%s %s),&index);",Temp->ArgName,Temp->Qualifier,Temp->ArgType);
			Temp=Temp->Next;
		}
		fprintf(Fp,"\n\tclose(sockfd);");
		fprintf(Fp,"\n\treturn(retval);");
		fprintf(Fp,"\n}\n");
		Top=Top->Next;Pid++;
	}
	WriteCopies(Fp);
	fclose(Fp);
	return 0;
}

int WriteSwitchCase(FILE *Fp,struct FnList *Fl)
{	int i=1;
	struct FnList *Top = Fl ;
	fprintf(Fp,"\n\t\t\tif(recv(new_fd,BUF,MAXDATASIZE,0)==-1){");
	fprintf(Fp,"\n\t\t\t\tperror(\"recv\");");
	fprintf(Fp,"\n\t\t\t\texit(1);\n\t\t\t}");
	fprintf(Fp,"\n\t\t\tindex=0;");
	fprintf(Fp,"\n\t\t\tcopy_from_buf(BUF,&procid,sizeof(int),&index);");
	fprintf(Fp,"\n\t\t\tswitch(procid){");
	while(Top!=NULL)
	{	fprintf(Fp,"\n\t\t\t\tcase %d:",i++);
		fprintf(Fp,"\n\t\t\t\t\t%s_rpc",(Top->Fn)->FnName);
		fprintf(Fp,"(BUF,&index);\n\t\t\t\t\tbreak;");
		Top=Top->Next;
	}
	fprintf(Fp,"\n\t\t\t}");
	return 1 ;
}

void WriteFunctionCall(FILE *Fp,struct FnStruct *Fn)
{	struct ArgNode *Temp=(Fn->Args).Head;
	fprintf(Fp,"%s(",Fn->FnName);
	while(Temp!=NULL)
	{	if((Temp->ArgMode == OUT || Temp->ArgMode == INOUT) && Temp->ArgNumDim == 0)
			fprintf(Fp,"&%s",Temp->ArgName) ;
		else
			fprintf(Fp,"%s",Temp->ArgName);
		Temp=Temp->Next;
		if(Temp!=NULL)
			fprintf(Fp,",");
	}
	fprintf(Fp,")");
}

int WriteMain(FILE *Fp,struct FnList *Fl)
{	fprintf(Fp,"\nmain(){\n");
	fprintf(Fp,"\n\t int new_fd;");
	fprintf(Fp,"\n\tint procid,index=0;");
	fprintf(Fp,"\n\tstruct sockaddr_in my_addr;");
	fprintf(Fp,"\n\tstruct sockaddr_in their_addr;");
	fprintf(Fp,"\n\tint sin_size;");
	fprintf(Fp,"\n\tchar BUF[MAXDATASIZE];");
	fprintf(Fp,"\n\tint yes=1;");	
	fprintf(Fp,"\n\tsignal(SIGQUIT,end);");
	fprintf(Fp,"\n\tif((sockfd=socket(AF_INET,SOCK_STREAM,0))==-1){");
	fprintf(Fp,"\n\t\tperror(\"socket\");");
	fprintf(Fp,"\n\t\texit(1);\n\t}");
	fprintf(Fp,"\n\tmy_addr.sin_family=AF_INET;");
	fprintf(Fp,"\n\tmy_addr.sin_port=htons(PORT);");
	fprintf(Fp,"\n\tmy_addr.sin_addr.s_addr=INADDR_ANY;");
	fprintf(Fp,"\n\tbzero(&(my_addr.sin_zero),8);");
	fprintf(Fp,"\n\tif(setsockopt(sockfd,SOL_SOCKET,SO_REUSEADDR,&yes,sizeof(int)) == -1) {") ;
	fprintf(Fp,"\n\t\tperror(\"setsockopt\");") ;
        fprintf(Fp,"\n\t\texit(1);\n\t}") ;
	fprintf(Fp,"\n\tif(bind(sockfd,(struct sockaddr*)&my_addr,sizeof(struct sockaddr))==-1){");
	fprintf(Fp,"\n\t\tperror(\"bind\");exit(1);\n\t}");
	fprintf(Fp,"\n\tif(listen(sockfd,BACKLOG)==-1){");
	fprintf(Fp,"\n\t\tperror(\"listen\");exit(1);}");
	fprintf(Fp,"\n\twhile(1){");
	fprintf(Fp,"\n\t\tsin_size=sizeof(struct   sockaddr_in);");
	fprintf(Fp,"\n\t\tif((new_fd=accept(sockfd,(struct sockaddr *)&their_addr,&sin_size))==-1){");
	fprintf(Fp,"\n\t\t\tperror(\"listen\");continue;}");
	fprintf(Fp,"\n\t\tprintf(\"server: got connection from %s\\n\",inet_ntoa(their_addr.sin_addr));","%s");
	fprintf(Fp,"\n\t\tif(!fork()){  /*child*/");
	WriteSwitchCase(Fp,Fl);
	fprintf(Fp,"\n\t\t\tif(send(new_fd,BUF,index,0)==-1){");
	fprintf(Fp,"\n\t\t\t\tperror(\"send\");\n\t\t\t\t\texit(1);");
	fprintf(Fp,"\n\t\t\t}");
	fprintf(Fp,"\n\t\t\tclose(new_fd);");
	fprintf(Fp,"\n\t\t\texit(0);");
	fprintf(Fp,"\n\t\t}");
	fprintf(Fp,"\n\t\tsignal(SIGCHLD,SIG_IGN);");
	fprintf(Fp,"\n\t\tclose(new_fd);");
	fprintf(Fp,"\n\t}");
	fprintf(Fp,"\n}\n");
	WriteCopies(Fp);
	return 1 ;
}

int CreateServer(char *Fn,struct FnList *Fl)
{	char *Fname;
	char *Tmp;
	FILE *Fp;
	int Ch,i ;
	struct ArgNode *Temp;
	struct FnList *Top = Fl ;
	Fname=(char *)malloc(sizeof(char)*strlen(Fn)+7);
	strcpy(Fname,Fn);
	if((Tmp=strstr(Fname,"."))!=NULL)
		Tmp[0]='\0';
	strcat(Fname,"_ser.c");
	if((Fp=fopen(Fname,"w+"))==NULL)
	{	perror("open");
		free(Fname);
		return(1);
	}
	free(Fname);
	fprintf(Fp,"/*******************************************************") ;
	fprintf(Fp,"\n*RPC server stub file created by RPCGEN");
	fprintf(Fp,"\n*\n*SOURCE IDL FILE : %s",Fn) ;
	fprintf(Fp,"\n*\n*Donot try to edit");
	fprintf(Fp,"\n*******************************************************/\n") ;
	fprintf(Fp,"#include<sys/socket.h>\n");
	fprintf(Fp,"#include<sys/types.h>\n");
	fprintf(Fp,"#include<errno.h>\n");
	fprintf(Fp,"#include<stdlib.h>\n");
	fprintf(Fp,"#include<string.h>\n");
	fprintf(Fp,"#include<netinet/in.h>\n");
	fprintf(Fp,"#include<stdio.h>\n");
	fprintf(Fp,"#include<sys/wait.h>\n");
	fprintf(Fp,"#include<signal.h>\n");
	fprintf(Fp,"\n#define PORT 3119");
	fprintf(Fp,"\n#define BACKLOG 10");
	fprintf(Fp,"\n#define MAXDATASIZE 1024");
	fprintf(Fp,"\nint sockfd;");
	fprintf(Fp,"\nvoid end(int num){") ;
	fprintf(Fp,"\n\tprintf(\"Exiting\");");
	fprintf(Fp,"\n\tshutdown(sockfd,SHUT_RDWR);");
	fprintf(Fp,"\n\texit(0);\n}");
	while(Top!=NULL)
	{	fprintf(Fp,"\n\n\n");
		WriteFunctionname(Fp,Top->Fn);
		fprintf(Fp,"(");
		fprintf(Fp,"char *BUF,int *index){\n");
		Temp=((Top->Fn)->Args).Head;
		while(Temp!=NULL)
		{	fprintf(Fp,"\t%s %s",Temp->Qualifier,Temp->ArgType);
			fprintf(Fp," ");
			fprintf(Fp,"%s",Temp->ArgName);
			if(Temp->ArgNumDim != 0)
			{	for(i = 0 ; i < Temp->ArgNumDim ; i++)
					fprintf(Fp,"[%d]",Temp->ArgDim[i]);
			}
			Temp=Temp->Next;
			fprintf(Fp,";\n");
		}
		fprintf(Fp,"\n\t%s retval;",(Top->Fn)->RetType);
		WriteArgcp2(Fp,Top->Fn);
		fprintf(Fp,"\n\tretval=");
		WriteFunctionCall(Fp,Top->Fn);
		fprintf(Fp,";");
		fprintf(Fp,"\n\t*index=0;");
		fprintf(Fp,"\n\tcopy_to_buf(BUF,(char *)&retval,sizeof(retval),index);");
		Temp=((Top->Fn)->Args).Head;
		while(Temp != NULL)
		{	if(Temp->ArgMode == INOUT || Temp->ArgMode == OUT)
				if(Temp->ArgNumDim != 0)
				{	int Size = 1,i ;	
					for(i = 0 ; i < Temp->ArgNumDim ; i++)
						Size *= Temp->ArgDim[i] ;
					fprintf(Fp,"\n\tcopy_to_buf(BUF,(char *)%s,sizeof(%s %s)*%d,index);",Temp->ArgName,Temp->Qualifier,Temp->ArgType,Size);
				}
				else
					fprintf(Fp,"\n\tcopy_to_buf(BUF,(char *)&%s,sizeof(%s %s),index);",Temp->ArgName,Temp->Qualifier,Temp->ArgType);
			Temp=Temp->Next;
		}
		fprintf(Fp,"\n\treturn (0);\n}");
		Top=Top->Next;
	}
	WriteMain(Fp,Fl);
	fclose(Fp);
	return(0);
}

int main(int argc,char *argv[])
{	struct ISLScanner S ;
	struct FnList *FL = NULL,*CurrNode = NULL ,*PrevNode = NULL ;
	struct FnStruct F1 ;
	char Line[MAX],*P ;
	int LineLen ;
	int ParserExitStatus = 0 ;
	FILE *InFile ;
	char fn[MAX] ;
	if(argc != 2)
		printf("Usage RPCGEN <filename> \n") ;
	else
	{     	if((InFile = fopen(argv[1],"r")) == NULL)
			printf("Error reading input file\n") ;
		else
		{	/*fseek(InFile,-1,SEEK_END) ;
				fputc('\n',InFile) ;
			fseek(InFile,0L,SEEK_SET) ;*/
			while(!feof(InFile))
			{	P = NULL ;
				fgets(Line,MAX,InFile) ;
				P = strstr(Line,"//") ;
				if(P != NULL)
					*P = '\0' ;
				LineLen = strlen(Line) ;
                		if(Line[LineLen-1] == '\n')
					Line[LineLen-1] = '\0' ;
				if(strlen(Line) <= 0)
					continue ;
				CurrNode = (struct FnList *)malloc(sizeof(struct FnList)) ;
				if(PrevNode != NULL)
					PrevNode->Next = CurrNode ;
				if(FL == NULL)
					FL = CurrNode ;
				CurrNode->Fn = (struct FnStruct *)malloc(sizeof(struct FnStruct)) ;
				InitFnStruct(CurrNode-> Fn) ;
				InitISLScanner(&S,Line) ;
				if(Function(&S,CurrNode->Fn) == 0)
				{	ParserExitStatus = 1 ;
					break ;
				}
				else if(FnSearch(FL,(CurrNode->Fn)->FnName) != CurrNode)
				{	printf("%s\nRedeclaration of function\n",Line) ;
					ParserExitStatus = 1 ;
					break ;
				}
				PrevNode = CurrNode ;
				CurrNode = CurrNode->Next ;
			}
			if(ParserExitStatus != 1)
			{	CreateClient(argv[1],FL) ;
				CreateServer(argv[1],FL) ;
			}
		}
	}
	return 0 ;
}
